from cached_property import cached_property
from colour import Color
from onegov.core.utils import safe_format_keys
from onegov.directory import DirectoryConfiguration
from onegov.directory import DirectoryZipArchive
from onegov.form import as_internal_id
from onegov.form import flatten_fieldsets
from onegov.form import Form
from onegov.form import merge_forms
from onegov.form import parse_formcode
from onegov.form.errors import FormError
from onegov.form.fields import IconField
from onegov.form.fields import UploadField
from onegov.form.filters import as_float
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import ValidFormDefinition
from onegov.form.validators import WhitelistedMimeType
from onegov.org import _
from onegov.org.forms.fields import HtmlField
from onegov.org.forms.generic import PaymentMethodForm
from onegov.org.theme.org_theme import user_options
from sqlalchemy.orm import object_session
from wtforms import BooleanField
from wtforms import DecimalField
from wtforms import RadioField
from wtforms import StringField
from wtforms import TextAreaField
from wtforms import ValidationError
from wtforms import validators
from wtforms_components import ColorField


class DirectoryBaseForm(Form):
    """ Form for directories. """

    title = StringField(
        label=_("Title"),
        fieldset=_("General"),
        validators=[validators.InputRequired()])

    lead = TextAreaField(
        label=_("Lead"),
        fieldset=_("General"),
        description=_("Describes what this directory is about"),
        render_kw={'rows': 4})

    text = HtmlField(
        label=_("Further Information"),
        fieldset=_("General"))

    structure = TextAreaField(
        label=_("Definition"),
        fieldset=_("General"),
        validators=[
            validators.InputRequired(),
            ValidFormDefinition(require_email_field=False)
        ],
        render_kw={'rows': 32, 'data-editor': 'form'})

    enable_map = RadioField(
        label=_("Coordinates"),
        fieldset=_("General"),
        choices=[
            (
                'no',
                _("Entries have no coordinates")
            ),
            (
                'entry',
                _("Coordinates are shown on each entry")
            ),
            (
                'everywhere',
                _("Coordinates are shown on the directory and on each entry")
            ),
        ],
        default='everywhere')

    title_format = StringField(
        label=_("Title-Format"),
        fieldset=_("Display"),
        validators=[validators.InputRequired()],
        render_kw={'class_': 'formcode-format'})

    lead_format = StringField(
        label=_("Lead-Format"),
        fieldset=_("Display"),
        render_kw={'class_': 'formcode-format'})

    content_fields = TextAreaField(
        label=_("Main view"),
        fieldset=_("Display"),
        render_kw={'class_': 'formcode-select'})

    contact_fields = TextAreaField(
        label=_("Address"),
        fieldset=_("Display"),
        render_kw={
            'class_': 'formcode-select',
            'data-fields-exclude': 'fileinput,radio,checkbox'
        })

    keyword_fields = TextAreaField(
        label=_("Filters"),
        fieldset=_("Display"),
        render_kw={
            'class_': 'formcode-select',
            'data-fields-include': 'radio,checkbox'
        })

    thumbnail = TextAreaField(
        label=_("Thumbnail"),
        fieldset=_("Display"),
        render_kw={
            'class_': 'formcode-select',
            'data-fields-include': 'fileinput'
        }
    )

    marker_icon = IconField(
        label=_("Icon"),
        fieldset=_("Marker"),
    )

    marker_color_type = RadioField(
        label=_("Marker Color"),
        fieldset=_("Marker"),
        choices=[
            ('default', _("Default")),
            ('custom', _("Custom"))
        ],
        default='default'
    )

    marker_color_value = ColorField(
        label=_("Color"),
        fieldset=_("Marker"),
        depends_on=('marker_color_type', 'custom')
    )

    order = RadioField(
        label=_("Order"),
        fieldset=_("Order"),
        choices=[
            ('by-title', _("By title")),
            ('by-format', _("By format"))
        ],
        default='by-title')

    order_format = StringField(
        label=_("Order-Format"),
        fieldset=_("Order"),
        render_kw={'class_': 'formcode-format'},
        validators=[validators.InputRequired()],
        depends_on=('order', 'by-format'))

    order_direction = RadioField(
        label=_("Direction"),
        fieldset=_("Order"),
        choices=[
            ('asc', _("Ascending")),
            ('desc', _("Descending"))
        ],
        default='asc'
    )

    link_pattern = StringField(
        label=_("Pattern"),
        fieldset=_("External Link"),
        render_kw={'class_': 'formcode-format'},
    )

    link_title = StringField(
        label=_("Title"),
        fieldset=_("External Link")
    )

    link_visible = BooleanField(
        label=_("Visible"),
        fieldset=_("External Link"),
        default=True
    )

    enable_submissions = BooleanField(
        label=_("Users may propose new entries"),
        fieldset=_("New entries"),
        default=True)

    guideline = HtmlField(
        label=_("Guideline"),
        fieldset=_("New entries"),
        depends_on=('enable_submissions', 'y'))

    price = RadioField(
        label=_("Price"),
        fieldset=_("New entries"),
        choices=[
            ('free', _("Free of charge")),
            ('paid', _("Paid"))
        ],
        default='free',
        depends_on=('enable_submissions', 'y'))

    price_per_submission = DecimalField(
        label=_("Price per submission"),
        fieldset=_("New entries"),
        filters=(as_float, ),
        validators=[validators.Optional()],
        depends_on=('enable_submissions', 'y', 'price', 'paid'))

    currency = StringField(
        label=_("Currency"),
        fieldset=_("New entries"),
        default="CHF",
        depends_on=('enable_submissions', 'y', 'price', 'paid'),
        validators=[validators.InputRequired()])

    @cached_property
    def known_field_ids(self):
        try:
            return {
                field.id for field in
                flatten_fieldsets(parse_formcode(self.structure.data))
            }
        except FormError:
            return None

    @cached_property
    def missing_fields(self):
        try:
            return self.configuration.missing_fields(self.structure.data)
        except FormError:
            return None

    def extract_field_ids(self, field):
        if not self.known_field_ids:
            return

        for line in field.data.splitlines():
            line = line.strip()

            if as_internal_id(line) in self.known_field_ids:
                yield line

    def validate_title_format(self, field):
        if self.missing_fields and 'title' in self.missing_fields:
            raise ValidationError(
                _("The following fields are unknown: ${fields}", mapping={
                    'fields': ', '.join(self.missing_fields['title'])
                }))

    def validate_lead_format(self, field):
        if self.missing_fields and 'lead' in self.missing_fields:
            raise ValidationError(
                _("The following fields are unknown: ${fields}", mapping={
                    'fields': ', '.join(self.missing_fields['lead'])
                }))

    def validate_thumbnail(self, field):
        if field.data and '\n' in field.data:
            raise ValidationError(
                _("Please select at most one thumbnail field")
            )

    @property
    def default_marker_color(self):
        return self.request.app.org.theme_options.get('primary-color')\
            or user_options['primary-color']

    @property
    def marker_color(self):
        if self.marker_color_value.data:
            return self.marker_color_value.data.get_hex()

    @marker_color.setter
    def marker_color(self, value):
        self.marker_color_value.data = Color(
            value or self.default_marker_color)

    @property
    def configuration(self):
        content_fields = list(self.extract_field_ids(self.content_fields))
        contact_fields = list(self.extract_field_ids(self.contact_fields))
        keyword_fields = list(self.extract_field_ids(self.keyword_fields))

        order_format = self.data[
            self.order.data == 'by-title' and 'title_format' or 'order_format'
        ]

        return DirectoryConfiguration(
            title=self.title_format.data,
            lead=self.lead_format.data,
            order=safe_format_keys(order_format),
            keywords=keyword_fields,
            searchable=content_fields + contact_fields,
            display={
                'content': content_fields,
                'contact': contact_fields
            },
            direction=self.order_direction.data,
            link_pattern=self.link_pattern.data,
            link_title=self.link_title.data,
            link_visible=self.link_visible.data,
            thumbnail=self.thumbnail.data and self.thumbnail.data.split()[0]
        )

    @configuration.setter
    def configuration(self, cfg):
        self.title_format.data = cfg.title
        self.lead_format.data = cfg.lead or ''
        self.content_fields.data = '\n'.join(cfg.display.get('content', ''))
        self.contact_fields.data = '\n'.join(cfg.display.get('contact', ''))
        self.keyword_fields.data = '\n'.join(cfg.keywords)
        self.order_direction = cfg.direction == 'desc' and 'desc' or 'asc'
        self.link_pattern.data = cfg.link_pattern
        self.link_title.data = cfg.link_title
        self.link_visible.data = cfg.link_visible
        self.thumbnail.data = cfg.thumbnail

        if safe_format_keys(cfg.title) == cfg.order:
            self.order.data = 'by-title'
        else:
            self.order.data = 'by-format'
            self.order_format.data = ''.join(f'[{key}]' for key in cfg.order)

    def populate_obj(self, obj):
        super().populate_obj(obj, exclude={
            'configuration',
            'order',
        })

        obj.configuration = self.configuration

        if self.marker_color_type.data == 'default':
            obj.marker_color = None
        else:
            obj.marker_color = self.marker_color

    def process_obj(self, obj):
        self.configuration = obj.configuration

        if obj.marker_color:
            self.marker_color_type.data = 'custom'
            self.marker_color = obj.marker_color
        else:
            self.marker_color_type.data = 'default'
            self.marker_color = self.default_marker_color


class DirectoryForm(merge_forms(DirectoryBaseForm, PaymentMethodForm)):

    payment_method_args = PaymentMethodForm.payment_method.kwargs.copy()
    payment_method_args['fieldset'] = _("New entries")
    payment_method_args['depends_on'] = (
        'enable_submissions', 'y', 'price', 'paid')

    payment_method = RadioField(**payment_method_args)


class DirectoryImportForm(Form):

    import_config = RadioField(
        label=_("Apply directory configuration"),
        choices=(
            ('yes', _("Yes, import configuration and entries")),
            ('no', _("No, only import entries"))
        ),
        default='no',
        validators=[validators.InputRequired()]
    )

    mode = RadioField(
        label=_("Mode"),
        choices=(
            ('new', _("Only import new entries")),
            ('replace', _("Replace all entries")),
        ),
        default='new',
        validators=[validators.InputRequired()]
    )

    zip_file = UploadField(
        label=_("Import"),
        validators=[
            validators.DataRequired(),
            WhitelistedMimeType({
                'application/zip',
                'application/octet-stream'
            }),
            FileSizeLimit(25 * 1024 * 1024)
        ],
        render_kw=dict(force_simple=True)
    )

    def run_import(self, target):
        session = object_session(target)

        count = 0

        def count_entry(entry):
            nonlocal count
            count += 1

        if self.mode.data == 'replace':
            for existing in target.entries:
                session.delete(existing)

            target.entries.clear()
            session.flush()

        archive = DirectoryZipArchive.from_buffer(self.zip_file.file)
        archive.read(
            target=target,
            skip_existing=True,
            limit=100,
            apply_metadata=self.import_config.data == 'yes',
            after_import=count_entry
        )

        return count



import re

from lxml import etree
from onegov.form import Form
from onegov.form import with_options
from onegov.form.fields import MultiCheckboxField, TagsField
from onegov.form.validators import Stdnum
from onegov.gis import CoordinatesField
from onegov.org import _
from onegov.org.forms.fields import HtmlField
from onegov.org.homepage_widgets import (
    transform_homepage_structure,
    XML_LINE_OFFSET
)
from onegov.org.theme import user_options
from wtforms import HiddenField, StringField, TextAreaField, RadioField
from wtforms import ValidationError
from wtforms import validators
from wtforms.fields.html5 import EmailField, URLField
from wtforms_components import ColorField


ERROR_LINE_RE = re.compile(r'line ([0-9]+)')


class SettingsForm(Form):
    """ Defines the settings form for onegov org. """

    name = StringField(
        label=_("Name"),
        validators=[validators.InputRequired()],
        fieldset=_("General")
    )
    logo_url = StringField(
        label=_("Logo"),
        description=_("URL pointing to the logo"),
        fieldset=_("General"),
        render_kw={'class_': 'image-url'}
    )
    square_logo_url = StringField(
        label=_("Logo (Square)"),
        description=_("URL pointing to the logo"),
        fieldset=_("General"),
        render_kw={'class_': 'image-url'}
    )
    reply_to = EmailField(
        _("E-Mail Reply Address"), [validators.InputRequired()],
        description=_("Replies to automated e-mails go to this address."),
        fieldset=_("General")
    )
    primary_color = ColorField(
        label=_("Primary Color"),
        fieldset=_("General")
    )
    locales = RadioField(
        label=_("Languages"),
        fieldset=_("General"),
        choices=(
            ('de_CH', _("German")),
            ('fr_CH', _("French"))
        ),
        validators=[validators.InputRequired()]
    )
    bank_account = StringField(
        label=_("Bank Account (IBAN)"),
        fieldset=_("Payment"),
        validators=[Stdnum(format='iban')]
    )
    bank_beneficiary = StringField(
        label=_("Beneficiary"),
        fieldset=_("Payment"),
    )
    bank_payment_order_type = RadioField(
        label=_("Payment Order"),
        fieldset=_("Payment"),
        choices=[
            ('basic', _("Basic")),
            ('esr', _("With reference number (ESR)")),
        ],
        default='basic'
    )
    bank_esr_participant_number = StringField(
        label=_("ESR participant number"),
        fieldset=_("Payment"),
        validators=[validators.InputRequired()],
        depends_on=('bank_payment_order_type', 'esr')
    )
    contact = TextAreaField(
        label=_("Contact"),
        description=_("The address and phone number of the municipality"),
        render_kw={'rows': 8},
        fieldset=_("Information")
    )
    contact_url = URLField(
        label=_("Contact Link"),
        description=_("URL pointing to a contact page"),
        fieldset=_("Information"),
        render_kw={'class_': 'internal-url'}
    )
    opening_hours = TextAreaField(
        label=_("Opening Hours"),
        description=_("The opening hours of the municipality"),
        render_kw={'rows': 8},
        fieldset=_("Information")
    )
    opening_hours_url = URLField(
        label=_("Opening Hours Link"),
        description=_("URL pointing to an opening hours page"),
        fieldset=_("Information"),
        render_kw={'class_': 'internal-url'}
    )
    facebook_url = URLField(
        label=_("Facebook"),
        description=_("URL pointing to the facebook site"),
        fieldset=_("Social Media")
    )
    twitter_url = URLField(
        label=_("Twitter"),
        description=_("URL pointing to the twitter site"),
        fieldset=_("Social Media")
    )
    homepage_image_1 = StringField(
        label=_("Homepage Image #1"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_2 = StringField(
        label=_("Homepage Image #2"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_3 = StringField(
        label=_("Homepage Image #3"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_4 = StringField(
        label=_("Homepage Image #4"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_5 = StringField(
        label=_("Homepage Image #5"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_image_6 = StringField(
        label=_("Homepage Image #6"),
        render_kw={'class_': 'image-url'},
        fieldset=_("Homepage")
    )
    homepage_cover = HtmlField(
        label=_("Homepage Cover"),
        render_kw={'rows': 10},
        fieldset=_("Homepage")
    )
    homepage_structure = TextAreaField(
        label=_("Homepage Structure (for advanced users only)"),
        description=_("The structure of the homepage"),
        render_kw={'rows': 32, 'data-editor': 'xml'},
        fieldset=_("Homepage")
    )

    # see homepage.py
    redirect_homepage_to = RadioField(
        label=_("Homepage redirect"),
        fieldset=_("Homepage"),
        default='no',
        choices=[
            ('no', _("No")),
            ('directories', _("Yes, to directories")),
            ('events', _("Yes, to events")),
            ('forms', _("Yes, to forms")),
            ('publications', _("Yes, to publications")),
            ('reservations', _("Yes, to reservations")),
        ],
    )

    hidden_people_fields = MultiCheckboxField(
        label=_("Hide these fields for non-logged-in users"),
        fieldset=_("People"),
        choices=[
            ('salutation', _("Salutation")),
            ('academic_title', _("Academic Title")),
            ('born', _("Born")),
            ('profession', _("Profession")),
            ('political_party', _("Political Party")),
            ('email', _("E-Mail")),
            ('phone', _("Phone")),
            ('phone_direct', _("Direct Phone Number")),
            ('website', _("Website")),
            ('address', _("Address")),
            ('notes', _("Notes")),
        ],
    )

    event_locations = TagsField(
        label=_("Values of the location filter"),
        fieldset=_("Events"),
    )

    default_map_view = CoordinatesField(
        label=_("The default map view. This should show the whole town"),
        render_kw={
            'data-map-type': 'crosshair'
        },
        fieldset=_("Maps")
    )

    geo_provider = RadioField(
        label=_("Geo provider"),
        fieldset=_("Maps"),
        default='geo-mapbox',
        choices=[
            ('geo-mapbox', _("Mapbox (Default)")),
            ('geo-vermessungsamt-winterthur', "Vermessungsamt Winterthur"),
            ('geo-zugmap-luftbild', "ZugMap Luftbild"),
            ('geo-zugmap-ortsplan', "ZugMap Ortsplan"),
        ]
    )

    analytics_code = TextAreaField(
        label=_("Analytics Code"),
        description=_("JavaScript for web statistics support"),
        render_kw={'rows': 10},
        fieldset=_("Advanced")
    )
    # the footer height is determined by javascript, see org.scss and
    # common.js for more information (search for footer)
    footer_height = HiddenField()

    def on_request(self):
        self.request.include('tags-input')

    def validate_homepage_structure(self, field):
        if field.data:
            try:
                transform_homepage_structure(self.request.app, field.data)
            except etree.XMLSyntaxError as e:
                correct_line = e.position[0] - XML_LINE_OFFSET

                correct_msg = 'line {}'.format(correct_line)
                correct_msg = ERROR_LINE_RE.sub(correct_msg, e.msg)

                field.widget = with_options(
                    field.widget, **{'data-highlight-line': correct_line}
                )

                raise ValidationError(correct_msg)

    @property
    def theme_options(self):
        try:
            primary_color = self.primary_color.data.get_hex()
        except AttributeError:
            primary_color = user_options['primary-color']

        options = {
            'primary-color': primary_color,
            'footer-height': self.footer_height.data
        }

        # set the images only if provided
        for i in range(1, 7):
            image = getattr(self, 'homepage_image_{}'.format(i)).data

            if not image:
                continue

            options['tile-image-{}'.format(i)] = '"{}"'.format(image)

        # override the options using the default vaules if no value was given
        for key in options:
            if not options[key]:
                options[key] = user_options[key]

        return options

    def ensure_beneificary_if_bank_account(self):
        if self.bank_account.data and not self.bank_beneficiary.data:
            self.bank_beneficiary.errors.append(_(
                "A beneficiary is required if a bank account is given."
            ))
            return False

    @theme_options.setter
    def theme_options(self, options):
        self.primary_color.data = options.get('primary-color')
        self.footer_height.data = options.get('footer-height')
        self.homepage_image_1.data = options.get('tile-image-1', '').strip('"')
        self.homepage_image_2.data = options.get('tile-image-2', '').strip('"')
        self.homepage_image_3.data = options.get('tile-image-3', '').strip('"')
        self.homepage_image_4.data = options.get('tile-image-4', '').strip('"')
        self.homepage_image_5.data = options.get('tile-image-5', '').strip('"')
        self.homepage_image_6.data = options.get('tile-image-6', '').strip('"')



import morepath

from onegov.core.security import Private
from onegov.core.utils import normalize_for_url
from onegov.form import FormCollection, FormDefinition
from onegov.org import _, OrgApp
from onegov.org.elements import Link
from onegov.org.forms import FormDefinitionForm
from onegov.org.layout import FormEditorLayout
from onegov.org.models import CustomFormDefinition
from webob import exc


def get_form_class(model, request):

    if isinstance(model, FormCollection):
        model = CustomFormDefinition()

    form_classes = {
        'builtin': FormDefinitionForm,
        'custom': FormDefinitionForm
    }

    return model.with_content_extensions(form_classes[model.type], request)


@OrgApp.form(model=FormCollection, name='new', template='form.pt',
             permission=Private, form=get_form_class)
def handle_new_definition(self, request, form):

    if form.submitted(request):

        if self.definitions.by_name(normalize_for_url(form.title.data)):
            request.alert(_("A form with this name already exists"))
        else:
            definition = self.definitions.add(
                title=form.title.data,
                definition=form.definition.data,
                type='custom'
            )
            form.populate_obj(definition)

            request.success(_("Added a new form"))
            return morepath.redirect(request.link(definition))

    layout = FormEditorLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Forms"), request.link(self)),
        Link(_("New Form"), request.link(self, name='new'))
    ]

    return {
        'layout': layout,
        'title': _("New Form"),
        'form': form,
        'form_width': 'large',
    }


@OrgApp.form(model=FormDefinition, template='form.pt', permission=Private,
             form=get_form_class, name='edit')
def handle_edit_definition(self, request, form):

    if form.submitted(request):
        form.populate_obj(self, exclude={'definition'})
        self.definition = form.definition.data

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))
    elif not request.POST:
        form.process(obj=self)

    collection = FormCollection(request.session)

    layout = FormEditorLayout(self, request)
    layout.breadcrumbs = [
        Link(_("Homepage"), layout.homepage_url),
        Link(_("Forms"), request.link(collection)),
        Link(self.title, request.link(self)),
        Link(_("Edit"), request.link(self, name='edit'))
    ]

    return {
        'layout': layout,
        'title': self.title,
        'form': form,
        'form_width': 'large',
    }


@OrgApp.view(model=FormDefinition, request_method='DELETE',
             permission=Private)
def delete_form_definition(self, request):

    request.assert_valid_csrf_token()

    if self.type != 'custom':
        raise exc.HTTPMethodNotAllowed()

    if self.has_submissions(with_state='complete'):
        raise exc.HTTPMethodNotAllowed()

    FormCollection(request.session).definitions.delete(
        self.name, with_submissions=False)



""" The newsletter view. """

import morepath

from collections import OrderedDict
from itertools import groupby
from onegov.core.security import Public, Private
from onegov.core.templates import render_template
from onegov.core.utils import linkify
from onegov.event import Occurrence, OccurrenceCollection
from onegov.newsletter import Newsletter
from onegov.newsletter import NewsletterCollection
from onegov.newsletter import Recipient
from onegov.newsletter import RecipientCollection
from onegov.newsletter.errors import AlreadyExistsError
from onegov.org import _, OrgApp
from onegov.org.forms import NewsletterForm
from onegov.org.forms import NewsletterSendForm
from onegov.org.forms import NewsletterTestForm
from onegov.org.forms import SignupForm
from onegov.org.layout import DefaultMailLayout
from onegov.org.layout import NewsletterLayout
from onegov.org.layout import RecipientLayout
from onegov.org.models import News
from sedate import utcnow
from sqlalchemy import desc
from sqlalchemy.orm import undefer
from string import Template


def get_newsletter_form(model, request):

    form = NewsletterForm

    query = request.session.query(News)
    query = query.filter(News.parent != None)
    query = query.order_by(desc(News.created))
    query = query.options(undefer('created'))
    form = form.with_news(request, query.all())

    query = OccurrenceCollection(request.session).query()
    subquery = query.with_entities(Occurrence.id)
    subquery = subquery.order_by(None)
    subquery = subquery.order_by(
        Occurrence.event_id,
        Occurrence.start
    )
    subquery = subquery.distinct(Occurrence.event_id).subquery()
    query.filter(Occurrence.id.in_(subquery))

    form = form.with_occurrences(request, query.all())

    return form


def news_by_newsletter(newsletter, request):
    news_ids = newsletter.content.get('news')

    if not news_ids:
        return None

    query = request.session.query(News)
    query = query.order_by(desc(News.created))
    query = query.options(undefer('created'))
    query = query.options(undefer('content'))
    query = query.filter(News.id.in_(news_ids))

    return request.exclude_invisible(query.all())


def occurrences_by_newsletter(newsletter, request):
    occurrence_ids = newsletter.content.get('occurrences')

    if not occurrence_ids:
        return None

    query = request.session.query(Occurrence)
    query = query.order_by(Occurrence.start, Occurrence.title)
    query = query.filter(Occurrence.id.in_(occurrence_ids))

    return query.all()


@OrgApp.form(model=NewsletterCollection, template='newsletter_collection.pt',
             permission=Public, form=SignupForm)
def handle_newsletters(self, request, form):

    if form.submitted(request):
        recipients = RecipientCollection(request.session)
        recipient = recipients.by_address(form.address.data)

        # do not show a specific error message if the user already signed up,
        # just pretend like everything worked correctly - if someone signed up
        # or not is private

        if not recipient:
            recipient = recipients.add(address=form.address.data)

            title = request.translate(
                _("Welcome to the ${org} Newsletter", mapping={
                    'org': request.app.org.title
                })
            )

            confirm_mail = render_template('mail_confirm.pt', request, {
                'layout': DefaultMailLayout(self, request),
                'newsletters': self,
                'subscription': recipient.subscription,
                'title': title
            })

            request.app.send_marketing_email(
                subject=title,
                receivers=(recipient.address, ),
                content=confirm_mail
            )

        request.success(_((
            "Success! We have sent a confirmation link to "
            "${address}, if we didn't send you one already."
        ), mapping={'address': form.address.data}))

    query = self.query()
    query = query.options(undefer(Newsletter.created))
    query = query.order_by(desc(Newsletter.created))

    # newsletters which were not sent yet are private
    if not request.is_manager:
        query = query.filter(Newsletter.sent != None)

    # the recipients count is only shown to logged in users
    if request.is_manager:
        recipients_count = RecipientCollection(self.session).query()\
            .filter(Recipient.confirmed == True)\
            .count()
    else:
        recipients_count = 0

    return {
        'form': form,
        'layout': NewsletterLayout(self, request),
        'newsletters': query.all(),
        'title': _("Newsletter"),
        'recipients_count': recipients_count
    }


@OrgApp.html(model=Newsletter, template='newsletter.pt', permission=Public)
def view_newsletter(self, request):

    return {
        'layout': NewsletterLayout(self, request),
        'newsletter': self,
        'news': news_by_newsletter(self, request),
        'occurrences': occurrences_by_newsletter(self, request),
        'title': self.title,
        'lead': linkify(self.lead),
    }


@OrgApp.html(model=RecipientCollection, template='recipients.pt',
             permission=Private)
def view_subscribers(self, request):

    # i18n:attributes translations do not support variables, so we need
    # to do this ourselves
    warning = request.translate(_("Do you really want to unsubscribe \"{}\"?"))

    recipients = self.query().order_by(Recipient.address).all()
    by_letter = OrderedDict()

    for key, values in groupby(recipients, key=lambda r: r.address[0].upper()):
        by_letter[key] = list(values)

    return {
        'layout': RecipientLayout(self, request),
        'title': _("Subscribers"),
        'by_letter': by_letter,
        'warning': warning,
    }


@OrgApp.form(model=NewsletterCollection, name='new', template='form.pt',
             permission=Public, form=get_newsletter_form)
def handle_new_newsletter(self, request, form):

    if form.submitted(request):
        try:
            newsletter = self.add(title=form.title.data, html='')
        except AlreadyExistsError:
            request.alert(_("A newsletter with this name already exists"))
        else:
            form.update_model(newsletter, request)

            request.success(_("Added a new newsletter"))
            return morepath.redirect(request.link(newsletter))

    return {
        'form': form,
        'layout': NewsletterLayout(self, request),
        'title': _("New Newsletter"),
        'size': 'large'
    }


@OrgApp.form(model=Newsletter, template='form.pt', name='edit',
             permission=Private, form=get_newsletter_form)
def edit_newsletter(self, request, form):

    if form.submitted(request):
        form.update_model(self, request)

        request.success(_("Your changes were saved"))
        return morepath.redirect(request.link(self))

    elif request.method == 'GET':
        form.apply_model(self)

    return {
        'layout': NewsletterLayout(self, request),
        'form': form,
        'title': _("Edit Newsletter"),
        'size': 'large'
    }


@OrgApp.view(model=Newsletter, request_method='DELETE', permission=Private)
def delete_page(self, request):
    request.assert_valid_csrf_token()

    NewsletterCollection(request.session).delete(self)
    request.success(_("The newsletter was deleted"))


def send_newsletter(request, newsletter, recipients, is_test=False):
    html = Template(render_template(
        'mail_newsletter.pt', request, {
            'layout': DefaultMailLayout(newsletter, request),
            'lead': linkify(newsletter.lead or '').replace('\n', '<br>'),
            'newsletter': newsletter,
            'title': newsletter.title,
            'unsubscribe': '$unsubscribe',
            'news': news_by_newsletter(newsletter, request),
            'occurrences': occurrences_by_newsletter(newsletter, request)
        }
    ))

    for count, recipient in enumerate(recipients, start=1):
        unsubscribe = request.link(recipient.subscription, 'unsubscribe')

        request.app.send_marketing_email(
            subject=newsletter.title,
            receivers=(recipient.address, ),
            content=html.substitute(unsubscribe=unsubscribe)
        )

        if not is_test and recipient not in newsletter.recipients:
            newsletter.recipients.append(recipient)

    if not is_test:
        newsletter.sent = newsletter.sent or utcnow()

    return count


@OrgApp.form(model=Newsletter, template='send_newsletter.pt', name='send',
             permission=Private, form=NewsletterSendForm)
def handle_send_newsletter(self, request, form):
    layout = NewsletterLayout(self, request)

    open_recipients = self.open_recipients

    if form.submitted(request):
        if form.send.data == 'now':
            sent = send_newsletter(request, self, open_recipients)

            request.success(_('Sent "${title}" to ${n} recipients', mapping={
                'title': self.title,
                'n': sent
            }))
        elif form.send.data == 'specify':
            self.scheduled = form.time.data

            request.success(
                _('Scheduled "${title}" to be sent on ${date}', mapping={
                    'title': self.title,
                    'date': layout.format_date(self.scheduled, 'datetime_long')
                })
            )
        else:
            raise NotImplementedError()

        return morepath.redirect(request.link(self))

    return {
        'layout': layout,
        'form': form,
        'title': self.title,
        'newsletter': self,
        'previous_recipients': self.recipients,
        'open_recipients': open_recipients,
    }


@OrgApp.form(model=Newsletter, template='form.pt', name='test',
             permission=Private, form=NewsletterTestForm.build)
def handle_test_newsletter(self, request, form):
    layout = NewsletterLayout(self, request)

    if form.submitted(request):
        send_newsletter(request, self, (form.recipient, ), is_test=True)

        request.success(_('Sent "${title}" to ${recipient}', mapping={
            'title': self.title,
            'recipient': form.recipient.address
        }))

        return morepath.redirect(request.link(self))

    return {
        'layout': layout,
        'form': form,
        'title': self.title,
        'helptext': _(
            "Sends a test newsletter to the given address"
        )
    }



import sedate

from datetime import datetime
from onegov.core.collection import GenericCollection
from onegov.file import File
from sqlalchemy import and_, text


class PublicationCollection(GenericCollection):

    def __init__(self, session, year=None):
        super().__init__(session)
        self.year = year

    @property
    def model_class(self):
        return File

    def query(self):
        query = super().query().filter_by(published=True, signed=True).filter(
            text("reference->>'content_type' = :content_type").bindparams(
                content_type='application/pdf'
            )
        )

        if self.year:
            s = sedate.replace_timezone(datetime(self.year, 1, 1), 'UTC')
            e = sedate.replace_timezone(datetime(self.year + 1, 1, 1), 'UTC')

            query = query.filter(and_(s <= File.created, File.created < e))

        return query

    def for_year(self, year):
        return self.__class__(self.session, year)

    def first_year(self, timezone):
        query = self.for_year(None).query()\
            .with_entities(File.created)\
            .order_by(File.created)

        first_record = query.first()

        if first_record:
            return sedate.to_timezone(first_record.created, timezone).year
